/* $Id: entrytable.cpp,v 1.3 2001/02/12 11:21:10 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <fstream>
#include <string>
#include <deque>
#include <map>
#include <stdio.h>

#include "entrytable.hpp"
#include "nib_io.hpp"
#include "kasm.hpp"


bool
read_entry_table(string filename, Entry_Table& tbl)
{
  ifstream fIn(filename.c_str());
  if (!fIn)
    return false;

  string line;
  char name_buf[BUFSIZE];
  unsigned long address;
  while (!fIn.eof()) {
    getline(fIn, line, '\n');

    if (sscanf(line.c_str(), "=%s EQU #%lx", name_buf, &address) != 2)
      continue;

    string key(name_buf);
    Entry* entry;
    if (key[0] == '~')          // Rompointer entry
      entry = new Romptr(address);
    else if (key[0] == '^')     // Flashpointer entry
      entry = new Flashptr(address);
    else
      entry = new Pointer(address);

    tbl[key] = entry;
  }
  fIn.close();
  
  return true;
}
