/* $Id: entrytable.hpp,v 1.4 2001/01/26 10:39:04 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */


#ifndef __ENTRYTABLE_HPP
#define __ENTRYTABLE_HPP


#define T_POINTER  0
#define T_ROMPTR   1
#define T_FLASHPTR 2

#define BUFSIZE 25


// The data types
class Entry          // Just for the other classes have a common ancestor
{
public:
  virtual int type() const = 0;
  virtual unsigned long get_address() const = 0;
};


class Pointer: public Entry 
{
public:
  Pointer(unsigned long address);

  int type() const;
  unsigned long get_address() const;
  
private:
  unsigned long address_;
};


class Romptr : public Entry
{
public:
  Romptr(unsigned long packed);

  int type() const;
  unsigned long get_address() const;
  unsigned long lib_num() const;
  unsigned long cmd_num() const;

private:
  unsigned long lib_;
  unsigned long cmd_;
};


class Flashptr : public Entry
{
public:
  Flashptr(unsigned long packed);

  int type() const;
  unsigned long get_address() const;
  unsigned long bank_num() const;
  unsigned long cmd_num() const;

private:
  unsigned long bank_;
  unsigned long cmd_;
};


inline
Pointer::Pointer(unsigned long address)
  : address_(address) { }


inline
int
Pointer::type() const
{
  return T_POINTER;
}


inline
unsigned long
Pointer::get_address() const
{
  return address_;
}


inline
Romptr::Romptr(unsigned long packed)
{
  lib_ = packed & 0xFFF;
  cmd_ = (packed >> 12) & 0xFFF;
}


inline
int
Romptr::type() const
{
  return T_ROMPTR;
}


inline
unsigned long
Romptr::get_address() const
{
  return ((cmd_ << 12) | lib_);
}


inline
unsigned long
Romptr::lib_num() const
{
  return lib_;
}


inline
unsigned long
Romptr::cmd_num() const
{
  return cmd_;
}


inline
Flashptr::Flashptr(unsigned long packed) 
{
  bank_ = packed & 0xFFF;
  cmd_ = (packed >> 12) & 0xFFFF;
}


inline
int
Flashptr::type() const
{
  return T_FLASHPTR;
}


inline
unsigned long
Flashptr::get_address() const
{
  return ((cmd_ << 12) | bank_);
}


inline
unsigned long
Flashptr::bank_num() const
{
  return bank_;
}


inline
unsigned long
Flashptr::cmd_num() const
{
  return cmd_;
}



// Type definitions
typedef map<string, Entry*> Entry_Table;
typedef Entry_Table::iterator Entry_Table_iter;


// Function prototypes
bool read_entry_table(string filename, Entry_Table& tbl);


#endif // __ENTRYTABLE_HPP
  
