/* $Id: kasm.hpp,v 1.53 2001/03/06 00:50:33 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */


#ifndef __KASM_HPP
#define __KASM_HPP


#define COPYRIGHT PACKAGE " v" VERSION ", Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)"

#define HOMEPAGE "kasm homepage is at http://k-asm.sourceforge.net\n"

#define HELPTEXT                                                           \
"\n"                                                                       \
"Usage: kasm [options] file\n\n"                                           \
"  -8, --hp48              Create a HP48 file" HP48_DEF_HLP "\n"           \
"  -9, --hp49              Create a HP49 file" HP49_DEF_HLP "\n"           \
"  -h, --header header     Header added to the generated object\n"         \
"  -d, --debug             Outputs debugging information for the parser\n" \
"  -o, --output file       Specifies alternate output file\n"              \
"  -t, --entry-table file  Specifies the entry table file to use\n"        \
"  -w, --no-warnings       Supresses warnings\n"                           \
"      --version           Displays version information and exits\n"       \
"      --help              Displays help information and exits\n"

/* Data structures */
struct s_equivtable 
{
  char mnem[20];
  unsigned long opcode;
};

enum { NONE, SKELSE, SKEC, SKENC, SKLSE };

typedef struct s_equivtable Equivtable;

struct s_block_info 
{
  unsigned long start;
  unsigned long end;
  int skelse_type;
  unsigned long else_block;
};

typedef struct s_block_info Block_Info;


/* Global variables */
extern unsigned long pc;              /* Program counter */
extern Entry_Table file_entries;      /* Entries defined in the source file */
extern Entry_Table global_entries;    /* Entries defined in the entry file */
extern bool firstpass;                /* Are we still in the first pass? */
extern unsigned long last_test_opcode; /* Opcode of last test instruction */
extern unsigned long last_test_rev;   /* Opcode of oposite of last test */
extern int last_test_len;             /* Length of last test opcode */
extern bool use_reversed;             /* The last test should be reversed? */
extern nibfstream fOut;               /* Output file */
extern bool ishp49;                   /* Is the output for the HP49? */
extern string filename;               /* Name of file being parsed. */
extern deque<unsigned long> code_lens; /* Lengths of code objects. */
extern unsigned long code_begin;      /* Start address of last code ob. */
extern unsigned long last_closed_block; /* Used in SKELSE and similars. */
extern unsigned long blk_count;       /* Number of current block. */
extern deque<Block_Info> blk_info;    /* Information about blocks blocks. */
extern deque<unsigned long> blk_curr; /* Current open block numbers. */


/* Function prototypes */
void mnemaddsub(char *mnem, char *field);
void mnemaddsubcon(char *mnem, char *field, long n);
void mnemzeroetal(char *mnem, char *field);
void mnemlogical(char *mnem, char *field);
void mnemsrb(long reg, char *field);
void mnemacld(char *mnem, long n);
void mnemmemoryf(char *mnem, char *field);
void mnemmemoryn(char *mnem, long size);
void mnembit(long opcode, long n);
void mnemdatregadd(char *mnem, long n);
void mnemdatregld(char *mnem, unsigned long n);
void mnemscratchreg(long op, char *field);
void mnempld(long n);
void mnemcpasn(long opcode, long n);
void mnemsetst(long opcode, long n);
void mnemgoto(char *mnem, long n);
void mnemtest1(char *mnem, char *field);
void mnemtest2(char *mnem, char *field);
void mnemtest3(long opcode, long n);
void mnemtest5(long opcode, long n);
void goyes(long n);
void mnemnofield2(char *mnem);
void mnemnofield3(char *mnem);
void mnemnofield4(char *mnem);
void mnemskip(bool isgosub);
void mnemskip_c(bool iscarry);
void mnemskip_l(bool isgosub);
void mnemskelse();
void mnemexitorup(long n, bool isexit);
void mnemexitorup_c(long n, bool isexit, bool iscarry);
void mnemstrobj(long prolog);
unsigned long reverse_test(unsigned long opcode);

void doptrbint(char *str, bool is_ptr);
void doId(char *name, bool is_id);
void dostring(char *str);
void dochar(unsigned char chr);
void doreal(char *str);
void doexreal(char *str);
void docmp(char *re_str, char *im_str);
void doexcmp(char *re_str, char *im_str);
void dotag(char *tag);
void doromptr(char *str1, char *str2);
void doflashptr(char *str1, char *str2);
void doacptr(char *str1, char *str2);
void dolenbody(char *prolog, char *len_str, char *body);
void donibb(char *len_str, char *body);
void dozint(char *str);

unsigned long lookup(Equivtable* table, int n, char *mnem);

unsigned long get_address_of(const string& entry);

void output(unsigned long opcode, unsigned int nibbles);
void outputreverse(unsigned long opcode, unsigned int nibbles);


#endif // __KASM_HPP
