/* -*-fundamental-*- */
/* $Id: kasm_lexer.ll,v 1.91 2001/03/13 13:13:44 ekalin Exp $ */
%{
/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string>
#include <deque>
#include <map>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "entrytable.hpp"
#include "nib_io.hpp"
#include "kasm.hpp"
#include "kasm_parser.h"


int dogosubvl(char *entry);
int dogotovl(char *entry);

deque<unsigned long> code_lens;
unsigned long code_begin;

static int previous_cond;

static bool seen_if = false;

static bool output_debug = true;

static deque<YY_BUFFER_STATE> include_buffers;
static deque<unsigned long>   include_line_ns;
static deque<string>          include_files;

static int string_len;
%}

%option 8bit
%option yylineno
%option never-interactive

%x CODE_S
%x STRING_S
%x ID_S
%x IGNORE_S

/* For picking up the name of the include files */
%x incl

%%


^\*.*$                    /* Ignore comments */
[ \t]+\([^)]*\)
^\([^)]*\)


<CODE_S>^[%;].*$          /* Ignore comments in assembly mode */
<CODE_S>[ \t]+[%;].*$


<INITIAL,CODE_S>"!IF48" {
        if (ishp49) {
                previous_cond = YY_START;
                BEGIN(IGNORE_S);
        } else
                seen_if = true;
}


<INITIAL,CODE_S>"!IF49" {
        if (!ishp49) {
                previous_cond = YY_START;
                BEGIN(IGNORE_S);
        } else
                seen_if = true;
}


<INITIAL,CODE_S>"!ENDIF" {
        if (!seen_if)
                yyerror("!ENDIF without !IF48 or !IF49");
        seen_if = false;
}


<INITIAL,CODE_S>"!OFF"                          /* Ignored. */
<INITIAL,CODE_S>"!WAROFF"                       /* Ignored. */
<INITIAL,CODE_S>"!NO CODE"                      /* Ignored. */
<INITIAL,CODE_S>"!COMPEX"                       /* Ignored. */  
<INITIAL,CODE_S>"!PATH+"[ \t]*[^ \t\n]+         /* Ignored. */


<INITIAL,CODE_S>"!DBGON"        output_debug = true;
<INITIAL,CODE_S>"!DBGOFF"       output_debug = false;


[ \t\n]+                  /* Skip whitespace */

FPTR2                     /* Ignore, entry type detection automatic. */


CODE[ \t]*$ {
        BEGIN(CODE_S);
        if (firstpass) {
                output(0, 10);
                code_begin = pc - 5;
        } else {
                unsigned long addr = get_address_of("DOCODE");
                if (addr == 0) {
                        yyerror("Cannot find address of code object prolog (DOCODE)", true);
                        yyterminate();
                }
                outputreverse(addr, 5);
                outputreverse(code_lens.back(), 5);
                code_lens.pop_back();
        }
        return CODE;
}


"::" {
        unsigned long addr = get_address_of("DOCOL");
        if (addr == 0) {
                yyerror("Cannot find address of secondary object prolog (DOCOL)", true);
                yyterminate();
        }
        outputreverse(addr, 5);
        return DOCOL;
}


";" {
        unsigned long addr = get_address_of("SEMI");
        if (addr == 0) {
                yyerror("Cannot find address of SEMI", true);
                yyterminate();
        }
        outputreverse(addr, 5);
        return SEMI;
}


"{" {
        unsigned long addr = get_address_of("DOLIST");
        if (addr == 0) {
                yyerror("Cannot find address of list prolog (DOLIST)", true);
                yyterminate();
        }
        outputreverse(addr, 5);
        return '{';
}


"}" {
        unsigned long addr = get_address_of("SEMI");
        if (addr == 0) {
                yyerror("Cannot find address of SEMI", true);
                yyterminate();
        }
        outputreverse(addr, 5);
        return '}';
}


PTR     return PTR;


ROMPTR  return ROMPTR;


FPTR    return FPTR;


ACPTR   return ACPTR;


HXS     {
        strcpy(yylval.str, "DOHXS");
        return LENBODYOB;
}
GROB    {
        strcpy(yylval.str, "DOGROB");
        return LENBODYOB;
}
LIBDAT  {
        strcpy(yylval.str, "DOEXT0");
        return LENBODYOB;
}
BAK  {
        strcpy(yylval.str, "DOBAK");
        return LENBODYOB;
}
LIB     {
        strcpy(yylval.str, "DOLIB");
        return LENBODYOB;
}
EXT1    {
        strcpy(yylval.str, "DOEXT1");
        return LENBODYOB;
}
EXT2    {
        strcpy(yylval.str, "DOEXT2");
        return LENBODYOB;
}
EXT3    {
        strcpy(yylval.str, "DOEXT3");
        return LENBODYOB;
}
EXT4    {
        strcpy(yylval.str, "DOEXT4");
        return LENBODYOB;
}
ARRAY   {
        strcpy(yylval.str, "DOARRY");
        return LENBODYOB;
}
LNKARRY {
        strcpy(yylval.str, "DOLNKARRY");
        return LENBODYOB;
}
CODE    {
        strcpy(yylval.str, "DOCODE");
        return LENBODYOB;
}


NIBB    return NIBB;


NIBHEX  return NIBBHEX;
NIBBHEX return NIBBHEX;


ZINT    return ZINT;


SYMBOL {
        unsigned long addr = get_address_of("DOSYMB");
        if (addr == 0) {
                yyerror("Cannot find address of symbolic object prolog (DOSYMB)", true);
                yyterminate();
        }
        outputreverse(addr, 5);
        return SYMBOL;
}


UNIT {
        unsigned long addr = get_address_of("DOEXT");
        if (addr == 0) {
                yyerror("Cannot find address of unit prolog (DOEXT)", true);
                yyterminate();
        }
        outputreverse(addr, 5);
        return UNIT;
}


MATRIX {
        unsigned long addr = get_address_of("DOMATRIX");
        if (addr == 0) {
                yyerror("Cannot find address of matrix prolog (DOMATRIX)", true);
                yyterminate();
        }
        outputreverse(addr, 5);
        return MATRIX;
}


TAG     {
        BEGIN(ID_S);
        string_len = 0;
        return TAG;
}


ID      {
        BEGIN(ID_S);
        string_len = 0;
        return ID;
}
LAM     {
        BEGIN(ID_S);
        string_len = 0;
        return LAM;
}


CHR[ \t]*. {
        yylval.chr = yytext[strlen(yytext) - 1];
        if (yylval.chr == '\\') {
                unsigned char c = yyinput();
                if (c >= 'A' && c <= 'F')
                        c = c - 'A' + 10;
                else if (c >= '0' && c <= '9')
                        c -= '0';
                else
                        yyerror("Hexa expected");
                yylval.chr = c << 4;
                c = yyinput();
                if (c >= 'A' && c <= 'F')
                        c = c - 'A' + 10;
                else if (c >= '0' && c <= '9')
                        c -= '0';
                else
                        yyerror("Hexa expected");
                yylval.chr |= c;
        }
        return CHR;
}


LABEL   return LABEL;


DEFINE|EQU|DC   return DEFINE;
EQUCP           return DCCP;


INCLUDE {
        previous_cond = INITIAL;
        BEGIN(incl);
}


\"      {
        BEGIN(STRING_S);
        string_len = 0;
}


"@"              |
<INITIAL><<EOF>> {
        if (seen_if)
                yyerror("!IF48 or !IF49 without !ENDIF");

        if (include_buffers.size() == 0) {
                if (firstpass) {
                        fseek(yyin, 0, SEEK_SET);
                        YY_FLUSH_BUFFER;
                        yylineno = -1;
                        BEGIN(INITIAL);
                }
                yyterminate();
        } else {
                yy_delete_buffer(YY_CURRENT_BUFFER);
                yy_switch_to_buffer(include_buffers.back());
                include_buffers.pop_back();

                yylineno = include_line_ns.back();
                include_line_ns.pop_back();
                filename = include_files.back();
                include_files.pop_back();
        }
}


"$"     return '$';


"#"     return '#';


"%"     return REAL;
"%%"    return EXREAL;
"C%"    return CMP;
"C%%"   return EXCMP;


[!#-~]+ {
        strcpy(yylval.str, yytext);
        return NAME;
}


.|\n    return *yytext;



<incl>[ \t]*            /* Ignore whitespace */
<incl>[^ \t\n]+ {       /* This is the file name */
        include_buffers.push_back(YY_CURRENT_BUFFER);

        if ((yyin = fopen(yytext, "r")) == 0) {
                string errormsg("Cannot open ");
                errormsg += string(yytext) + ": ";
                errormsg += strerror(errno);
                yyerror(errormsg.c_str());
                exit(1);
        }

        include_line_ns.push_back(yylineno);
        yylineno = 1;
        include_files.push_back(filename);
        filename = yytext;
        
        yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));

        BEGIN(previous_cond);
}


<STRING_S,ID_S>\\. {
        switch (yytext[1]) {
        case 'n':
                yylval.str[string_len++] = '\n';
                break;

        case 't':
                yylval.str[string_len++] = '\t';
                break;

        case '\"':
                yylval.str[string_len++] = '\"';
                break;

        case '\\':
                yylval.str[string_len++] = '\\';
                break;

        case ' ':
                yylval.str[string_len++] = ' ';
                break;

        default:
                if (isxdigit(yytext[1])) {
                        unsigned char c = yytext[1];
                        if (c >= 'A' && c <= 'F')
                                c = c - 'A' + 10;
                        else if (c >= '0' && c <= '9')
                                c -= '0';
                        yylval.str[string_len] = c << 4;
                        c = yyinput();
                        if (c >= 'A' && c <= 'F')
                                c = c - 'A' + 10;
                        else if (c >= '0' && c <= '9')
                                c -= '0';
                        else
                                yyerror("Hexa expected");
                        yylval.str[string_len++] |= c;
                } else {
                        yylval.str[string_len++] = '\\';
                        yylval.str[string_len++] = yytext[1];
                }
        }
}
<STRING_S>\"    {
        yylval.str[string_len] = '\0';
        BEGIN(INITIAL);
        return STRING;
}
<ID_S>[ \t\n]   {
        if (string_len != 0) {
                yylval.str[string_len] = '\0';
                BEGIN(INITIAL);
                return STRING;
        }
}
<STRING_S,ID_S>.        yylval.str[string_len++] = *yytext;



        /* The rules below are for CODE parsing. */
<CODE_S>{
[ \t]+                  /* Skip whitespace */


[ \t]+\*[A-Za-z0-9]+ |
^\*[A-Za-z0-9]+      {        /* A label */
        if (firstpass) {
                string stryytext(yytext);
                unsigned int pos = stryytext.find("*");
                string label = stryytext.substr(pos + 1);
                file_entries.insert(make_pair(label, new Pointer(pc)));
        }
}


ENDCODE {
        BEGIN(INITIAL);
        return ENDCODE;
}


DC|EQU|DEFINE           return DEFINE;


CON\([1-8]\) {
        yylval.number = yytext[4] - '0';
        return CON;
}
EXP\([1-8]\) {
        yylval.number = yytext[4] - '0';
        return CON;
}
\$\([1-8]\) {
        yylval.number = yytext[2] - '0';
        return CON;
}


NIBHEX[ \t]*[0-9A-F]+ {
        char *start = yytext + 6;
        while (isspace(*start))
                ++start;

        while (*start) {
                if (*start >= 'A')
                        output(*start - 'A' + 10, 1);
                else
                        output(*start - '0', 1);
                ++start;
        }
}


GOIN[AC] {
        yylval.number = (yytext[4] == 'A');
        return GOINAC;
}


GOIN[2-5] {
        yylval.number = yytext[4] - '0';
        return GOINN;
}
G[2-5] {
        yylval.number = yytext[1] - '0';
        return GOINN;
}


CP= {
        char c;
        while (true) {
                c = yyinput();
                if (isxdigit(c)) {
                        unput(c);
                        unput('$');
                        return CP;
                }
                if (!isspace(c)) {
                        unput(c);
                        return CP;
                }
        }
}


DCCP {
        char c;
        while (true) {
                c = yyinput();
                if (isxdigit(c)) {
                        unput(c);
                        unput('$');
                        return DCCP;
                }
                if (!isspace(c)) {
                        unput(c);
                        return DCCP;
                }
        }
}



GOYES                   return GOYES;


RTNYES|RTY              return RTNYES;


RTI|RTN(CC|SC|SXM)?   | 
SET(HEX|DEC)          |
RSTK=C|C=RSTK         |
CLRST|C=ST|ST=C|CSTEX |
(P=)?P[-+]1           {
        strcpy(yylval.str, yytext);
        return MNEMNOFIELD2;
}

RTNN?C           |
NOP3             |
D[01]=[AC]S?     |
[AC]D[01](EX|XS) |
OUT=CS?          |
[AC]=IN          |
UNCNFG|CONFIG    |
C=ID             |
SHUTDN           |
C\+P\+1          |
RESET            |
BUSCC            |
SREQ\?           |
[A-D]S[LR]C      |
(XM|S[BR]|MP)=0  |
CLRHST           {
        strcpy(yylval.str, yytext);
        return MNEMNOFIELD3;
}
"C=C+P+1"        {
        strcpy(yylval.str, "C+P+1");
        return MNEMNOFIELD3;
}


NOP4            |
BUSC[BD]        |
INT(ON|OFF)     |
PC=\([AC]\)     |
PC=[AC]|[AC]=PC |
[AC]PCEX        {
        strcpy(yylval.str, yytext);
        return MNEMNOFIELD4;
}


NOP5 {
        yylval.number = 0x64000;
        return MNEMNOFIELD5;
}
RSI {
        yylval.number = 0x80810;
        return MNEMNOFIELD5;
}


[A-D]=[A-D][-+][A-D] {
        strcpy(yylval.str, yytext);
        return MNEMADDSUB;
}
[A-D][+-][A-D] {
        yylval.str[0] = yytext[0];
        yylval.str[1] = '=';
        strcpy(yylval.str + 2, yytext);
        return MNEMADDSUB;
}


[A-D]=[A-D][+-] {
        strcpy(yylval.str, yytext);
        return MNEMADDSUBCON;
}
[A-D][+-] {
        yylval.str[0] = yytext[0];
        yylval.str[1] = '=';
        strcpy(yylval.str + 2, yytext);
        return MNEMADDSUBCON;
}


[A-D]=[0A-D]      |
[A-D][A-D]EX      |
[A-D]S[LR]        |
[A-D]=-[A-D](-1)? {
        strcpy(yylval.str, yytext);
        return MNEMZEROETAL;
}

[A-D]=[A-D][&!][A-D] {
        strcpy(yylval.str, yytext);
        return MNEMLOGICAL;
}
[A-D][&!][A-D] {
        yylval.str[0] = yytext[0];
        yylval.str[1] = '=';
        strcpy(yylval.str + 2, yytext);
        return MNEMLOGICAL;
}


[A-D]SRB {
        yylval.number = yytext[0] - 'A';
        return MNEMSRB;
}        


L[AC][ \t]*#?[0-9A-F]+ {           /* Hex loading */
        char *end = strchr(yytext, '\0') - 1;
        char *start = yytext + 2;
        while (isspace(*start))
                ++start;

        if (*start == '#') {
                sprintf(yylval.str, "L%c(%d)", yytext[1], end - start + 1);
                yyless((int)(start - yytext));
                return MNEMACLD;
        }

        --start;

        int size = end - start - 1;
        if (size > 15) {
                yyerror("Constant too large");
                size = 15;
        }

        if (yytext[1] == 'C')
                output(0x30 | size, 2);
        else
                output(0x80820 | size, 5);

        while (end != start) {
                if (*end >= 'A')
                        output(*end - 'A' + 10, 1);
                else
                        output(*end - '0', 1);
                --end;
        }
}
L[AC]\((0?[0-9]|1[0-6])\) {
        strcpy(yylval.str, yytext);
        return MNEMACLD;
}
L[AC]ASC\([1-8]\) {
        if (yytext[1] == 'C')
                output(3, 1);
        else
                output(0x8082, 4);

        int size = yytext[6] - '0';
        output(2*size - 1, 1);

        char c;
        while (true) {
                c = yyinput();
                if (!isspace(c))
                        break;
        }
        
        for (int i = 0; i < size; ++i, c = yyinput())
                outputreverse((unsigned long)c, 2);
}
        

DAT[01]=[AC] |
[AC]=DAT[01] {
        strcpy(yylval.str, yytext);
        return MNEMMEMORY;
}


[AC]BIT=[01] {
        yylval.number = 0x80840 
                +0x40*(yytext[0] == 'C') + 0x10*(yytext[5] == '1');
        return MNEMBIT;
}


D[01]=D[01][-+] {
        strcpy(yylval.str, yytext);
        return MNEMDATREGADD;
}
D[01][-+] {                     /* Make initial Dx= part */
        strncpy(yylval.str, yytext, 2);
        yylval.str[2] = '=';
        strcpy(yylval.str + 3, yytext);
        return MNEMDATREGADD;
}


D[01]=\([245]\) {
        strcpy(yylval.str, yytext);
        return MNEMDATREGLD;
}
D[01]= {
        strcpy(yylval.str, yytext);
        char c;
        while (true) {
                c = yyinput();
                if (isxdigit(c)) {
                        unput(c);
                        unput('$');
                        return MNEMDATREGLD;
                }
                if (!isspace(c)) {
                        unput(c);
                        return MNEMDATREGLD;
                }
        }
}


R[0-4]=[AC] {
        yylval.number = (yytext[1] - '0') + 8*(yytext[3] == 'C');
        return MNEMSCRATCHREG;
}
[AC]=R[0-4] {
        yylval.number = 0x10 + (yytext[3] - '0') + 8*(yytext[0] == 'C');
        return MNEMSCRATCHREG;
}
[AC]R[0-4]EX {
        yylval.number = 0x20 + (yytext[2] - '0') + 8*(yytext[0] == 'C');
        return MNEMSCRATCHREG;
}


P=                      return MNEMPLD;


C=P {
        yylval.number = 0x80C0;
        return MNEMCPASN;
}
P=C {
        yylval.number = 0x80D0;
        return MNEMCPASN;
}
CPEX {
        yylval.number = 0x80F0;
        return MNEMCPASN;
}


ST=[01] {
        yylval.number = 0x840 + 0x10*(yytext[3] == '1');
        return MNEMSETST;
}
HST=0 {
        yylval.number = 0x820;
        return MNEMSETST;
}


GOTO|GOSUB|GOC|GONC|GOLONG|GOTOL|GOSUBL|GOVLNG|GOSBVL {
        strcpy(yylval.str, yytext);
        return MNEMGOTO;
}


GOTOC {
        yylval.number = 0x500;
        return MNEMGOTOC;
}
GOTONC {
        yylval.number = 0x400;
        return MNEMGOTOC;
}


\?[A-D][=#\x8D][A-D0] {
        strcpy(yylval.str, yytext);
        if (yylval.str[2] == '\x8D')
                yylval.str[2] = '#';
        return MNEMTEST1;
}
\?[A-D]\\=\/[A-D0] {
        yylval.str[0] = '?';
        yylval.str[1] = yytext[1];
        yylval.str[2] = '#';
        strcpy(yylval.str + 3, yytext + 5);
        return MNEMTEST1;
}


\?[A-D]>=?[A-D] |
\?[A-D]<=?[A-D] {
        strcpy(yylval.str, yytext);
        return MNEMTEST2;
}
\?[A-D]\\[<>]=[A-D] {
        yylval.str[0] = '?';
        yylval.str[1] = yytext[1];
        strcpy(yylval.str + 2, yytext + 3);
        return MNEMTEST2;
}
\?[A-D][\x89\x8A][A-D] {
        yylval.str[0] = '?';
        yylval.str[1] = yytext[1];
        yylval.str[2] = (yytext[2] == '\x89' ? '<' : '>');
        yylval.str[3] = '=';
        yylval.str[4] = yytext[3];
        yylval.str[5] = '\0';
        return MNEMTEST2;
}


\?ST=[01] {
        yylval.number = (yytext[4] == '0' ? 0x860 : 0x870);
        last_test_rev = (yytext[4] != '0' ? 0x860 : 0x870);
        return MNEMTEST3;
}
\?HST=0 {
        yylval.number = 0x830;
        last_test_rev = 0;
        return MNEMTEST3;
}
\?P[=#\x8B] {
        yylval.number = (yytext[2] == '=' ? 0x890 : 0x880);
        last_test_rev = (yytext[2] != '=' ? 0x890 : 0x880);
        return MNEMTEST3;
}
\?P\\=\/ {
        yylval.number = 0x880;
        last_test_rev = 0x890;
        return MNEMTEST3;
}


\?XM=0 {
        yylval.number = 0x831;
        last_test_rev = 0;
        return MNEMTEST4;
}
\?SB=0 {
        yylval.number = 0x832;
        last_test_rev = 0;
        return MNEMTEST4;
}
\?SR=0 {
        yylval.number = 0x834;
        last_test_rev = 0;
        return MNEMTEST4;
}
\?MP=0 {
        yylval.number = 0x838;
        last_test_rev = 0;
        return MNEMTEST4;
}


\?[AC]BIT=[01] {
        yylval.number = 0x80860 + 0x10*(yytext[6] == '1') 
                        + 0x40*(yytext[1] == 'C');
        last_test_rev = 0x80860 + 0x10*(yytext[6] != '1') 
                        + 0x40*(yytext[1] == 'C');
        return MNEMTEST5;
}


LOOP|RPL {
        // Equivalent to A=DAT0.A  D0+5  PC=(A)
        mnemmemoryf("A=DAT0", "A");             // A=DAT0.A
        mnemdatregadd("D0=D0+", 5);             // D0+5
        mnemnofield4("PC=(A)");                 // PC=(A)
}
SAVE                    return dogosubvl("SAVPTR");
LOADRPL                 return dogotovl("GETPTRLOOP");
LOAD                    return dogosubvl("GETPTR");
INTOFF2                 return dogosubvl("DisableIntr");
INTON2                  return dogosubvl("AllowIntr");
ERROR_C                 return dogotovl("ErrjmpC");
A=IN2                   return dogosubvl("AINRTN");
C=IN2                   return dogosubvl("CINRTN");
OUT=C=IN                return dogosubvl("OUTCINRTN");
RES\.STR                return dogosubvl("MAKE$N");
RES\.ROOM               return dogosubvl("GETTEMP");
RESRAM                  return dogosubvl("MAKERAM$");
SHRINK\$                return dogosubvl("Shrink$");
COPY<-|COPY\\<-|COPY\x8E|COPYDN         return dogosubvl("MOVEDOWN");
COPY->|COPY\\->|COPY\x8D|COPYUP         return dogosubvl("MOVEUP");
DISPKEY                 {
        if (output_debug && ishp49)
                return dogosubvl("\"DBUG.TOUCHE\"");
}
DISP                    {
        if (output_debug && ishp49)
                return dogosubvl("DBUG");
}
SRKLST                  return dogosubvl("Shrink$List");
SCREEN                  return dogosubvl("\"D0->Row1\"");
MENU                    return dogosubvl("\"D0->Sft1\"");
ZEROMEM                 return dogosubvl("WIPEOUT");
MULT\.A                 return dogosubvl("\"MULTB+A*C\"");
MULT                    return dogosubvl("MPY");
DIV\.A                  return dogosubvl("IntDiv");
DIV                     return dogosubvl("IDIV");
BEEP                    return dogosubvl("makebeep");


SKIP    return SKIP;
SKUB    return SKUB;
SK(IP)?C        return SKIPC;
SK(IP)?NC       return SKIPNC;
SKIPL   return SKIPL;
SKUBL   return SKUBL;
SKIPYES return SKIPYES;


SKELSE {
        if (firstpass) {
                blk_info[last_closed_block].skelse_type = SKELSE;
                blk_info[last_closed_block].else_block = blk_count;
                output(0, 4);
        }
}
SKEC {
        if (firstpass) {
                blk_info[last_closed_block].skelse_type = SKEC;
                blk_info[last_closed_block].else_block = blk_count; 
                output(0, 3);
       }
}
SKENC {
        if (firstpass) {
                blk_info[last_closed_block].skelse_type = SKENC;
                blk_info[last_closed_block].else_block = blk_count;
                output(0, 3);
        }
}
SKLSE {
        if (firstpass) {
                blk_info[last_closed_block].skelse_type = SKLSE;
                blk_info[last_closed_block].else_block = blk_count;
        }
}


EXIT[0-9]* {
        if (strlen(yytext) == 4)
                yylval.number = 0;
        else {
                yylval.number = atoi(yytext + 4) - 1;
                if (yylval.number < 0)
                        yyerror("Invalid number after EXIT");
        }
        return EXIT;
}
UP[0-9]* {
        if (strlen(yytext) == 2)
                yylval.number = 0;
        else {
                yylval.number = atoi(yytext + 2) - 1;
                if (yylval.number < 0)
                        yyerror("Invalid number after UP");
        }
        return UP;
}
EXITC[0-9]* {
        if (strlen(yytext) == 5)
                yylval.number = 0;
        else {
                yylval.number = atoi(yytext + 5) - 1;
                if (yylval.number < 0)
                        yyerror("Invalid number after EXITC");
        }
        return EXITC;
}
UPC[0-9]* {
        if (strlen(yytext) == 3)
                yylval.number = 0;
        else {
                yylval.number = atoi(yytext + 3) - 1;
                if (yylval.number < 0)
                        yyerror("Invalid number after UPC");
        }
        return UPC;
}
EXITNC[0-9]* {
        if (strlen(yytext) == 6)
                yylval.number = 0;
        else {
                yylval.number = atoi(yytext + 6) - 1;
                if (yylval.number < 0)
                        yyerror("Invalid number after EXITNC");
        }
        return EXITNC;
}
UPNC[0-9]* {
        if (strlen(yytext) == 4)
                yylval.number = 0;
        else {
                yylval.number = atoi(yytext + 4) - 1;
                if (yylval.number < 0)
                        yyerror("Invalid number after UPNC");
        }
        return UPNC;
}


STROBJ  return STROBJ;
STRING {
        /* Unput "DOCSTR " */
        unput(' '); unput('R'); unput('T'); unput('S'); unput('C');
        unput('O'); unput('D');
        return STROBJ;
}
CODE {
        /* Unput "DOODE " */
        unput(' '); unput('E'); unput('D'); unput('O'); unput('C');
        unput('O'); unput('D');
        return STROBJ;
}


A|B|M|P|S|W|WP|X|XS {
        strcpy(yylval.str, yytext);
        return FIELD;
}


[0-9]+ {
        yylval.number = strtol(yytext, NULL, 10);
        return NUMBER;
}
#[0-9]+ {
        yylval.number = strtol(yytext + 1, NULL, 10);
        return NUMBER;
}


\$[0-9A-F]+ {
        strcpy(yylval.str, yytext + 1);
        return HEX_NUMBER;
}


"$/"[0-9A-F]+ {
        strcpy(yylval.str, yytext + 2);
        return HEX_REV_NUMBER;
}


[A-Za-z0-9$]+ {
        strcpy(yylval.str, yytext);
        return NAME;
}
=[A-Za-z0-9$]+ {
        strcpy(yylval.str, yytext + 1);
        return NAME;
}
\"[^\"\n]+\" {
        strcpy(yylval.str, yytext + 1);
        yylval.str[strlen(yylval.str) - 1] = '\0';
        return NAME;
}


"(*)"   {
        yylval.number = pc;
        return NUMBER;
}


"<="|"\\<="|\x89        return LE;


">="|"\\>="|\x8A        return GE;


"<<"                    return LSHIFT;


">>"                    return RSHIFT;


"->"|"\\->"|\x8D        {
        use_reversed = true;
        return ARROW;
}


"\\=/"|\x8B             return '#';


"'" { 
        previous_cond = CODE_S;
        BEGIN(incl);
}


"{" {
        if (use_reversed || last_test_len) {
                if (last_test_len == 0 && use_reversed)
                        yyerror("Forbidden");
                else if (last_test_len != 0) {
                        int len = last_test_len;
                        last_test_len = 0;
                        if (use_reversed && last_test_rev == 0)
                                yyerror("This test cannot be reversed");
                        output(use_reversed ? last_test_rev : last_test_opcode, len);
                        if (firstpass)
                                output(0, 2);
                        else {
                                long address = blk_info[blk_count].end - pc;
                                switch(blk_info[blk_count].skelse_type) {
                                case SKELSE:
                                        address += 4;
                                        break;

                                case SKEC:
                                case SKENC:
                                        address += 3;
                                        break;
                                }
                                if (address < -(0x80) || address > 0x7F)
                                        yyerror("Jump too long");
                                outputreverse(address & 0xFF, 2);
                         }
                }
        }
        if (firstpass) {
                Block_Info blk_i;
                blk_i.start = pc;
                blk_i.skelse_type = NONE;
                blk_info.push_back(blk_i);
        }
        blk_curr.push_back(blk_count++);

        return '{';
}


"}" {   
        /* A great hack to allow the closing brace to be on the same
           line as a instruction. */
        unput('\xA3');  /* Arbitrary. */
        return '\n';
}
\xA3    {
        last_closed_block = blk_curr.back();

        if (firstpass)
                blk_info[blk_curr.back()].end = pc;
        blk_curr.pop_back();

        return '}';
}


"@"     |
<<EOF>> {
        if (include_buffers.size() == 0) {
                yyerror("Unfinished CODE object");
                yyterminate();
        } else {
                yy_delete_buffer(YY_CURRENT_BUFFER);
                yy_switch_to_buffer(include_buffers.back());
                include_buffers.pop_back();

                yylineno = include_line_ns.back();
                include_line_ns.pop_back();
                filename = include_files.back();
                include_files.pop_back();
        }
}


.|\n                    return *yytext;


}
        /* End of CODE rules. */


<IGNORE_S>"!ENDIF" {
        BEGIN(previous_cond);
}


<IGNORE_S>.|\n    /* Ignore everything */


<IGNORE_S><<EOF>> {
        yyerror("!IF48 or !IF49 without !ENDIF");
        yyterminate();
}


%%


int
dogosubvl(char *entry) {
        for (int i = strlen(entry) - 1; i >= 0; --i)
                unput(entry[i]);
        strcpy(yylval.str, "GOSBVL");
        return MNEMGOTO;
}


int
dogotovl(char *entry) {
        for (int i = strlen(entry) - 1; i >= 0; --i)
                unput(entry[i]);
        strcpy(yylval.str, "GOVLNG");
        return MNEMGOTO;
}
