/* -*-fundamental-*- */
%{
/* $Id: kasm_parser.yy,v 1.71 2001/03/13 13:13:44 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string>
#include <deque>
#include <map>
#include <stdio.h>

#include "entrytable.hpp"
#include "nib_io.hpp"
#include "kasm.hpp"


// File variables
static unsigned long constpointer = 0x80100;    // Used by DCCP
static long cpointerincr;                       // Used by : Inc ... :

// Global variables
bool use_reversed = false;
%}

%union {
  long number;
  char str[80];
  unsigned char chr;
}


%token          PTR
%token          ID
%token          LAM
%token <str>    STRING
%token <chr>    CHR
%token          REAL
%token          EXREAL
%token          CMP
%token          EXCMP
%token          DOCOL
%token          SEMI
%token          SYMBOL
%token          UNIT
%token          MATRIX
%token          TAG
%token          ROMPTR
%token          FPTR
%token          ACPTR
%token <str>    LENBODYOB
%token          NIBB
%token          NIBBHEX
%token          ZINT
%token          CODE
%token          ENDCODE

%token          LABEL
%token          DEFINE
%token          DCCP

%token <str>    NAME


/* Everything below is for CODE parsing. */
%token <number> NUMBER
%token <str>    HEX_NUMBER
%token <str>    HEX_REV_NUMBER

%token <str>    MNEMADDSUB
%token <str>    MNEMADDSUBCON
%token <str>    MNEMZEROETAL
%token <str>    MNEMLOGICAL
%token <number> MNEMSRB
%token <str>    MNEMACLD
%token <str>    MNEMMEMORY
%token <number> MNEMBIT
%token <str>    MNEMDATREGADD
%token <str>    MNEMDATREGLD
%token <str>    MNEMPLD
%token <number> MNEMSCRATCHREG
%token <number> MNEMCPASN
%token <number> MNEMSETST
%token <str>    MNEMGOTO
%token <number> MNEMGOTOC
%token <str>    MNEMTEST1
%token <str>    MNEMTEST2
%token <number> MNEMTEST3
%token <number> MNEMTEST4
%token <number> MNEMTEST5
%token <str>    MNEMNOFIELD2
%token <str>    MNEMNOFIELD3
%token <str>    MNEMNOFIELD4
%token <number> MNEMNOFIELD5
%token          SKIP
%token          SKUB
%token          SKIPC
%token          SKIPNC
%token          SKIPL
%token          SKUBL
%token          SKIPYES
%token <number> EXIT
%token <number> UP
%token <number> EXITC
%token <number> UPC
%token <number> EXITNC
%token <number> UPNC
%token          STROBJ

%token <str>    FIELD

%token <number> CON
%token <number> GOINAC
%token <number> GOINN
%token          CP
%token          GOYES
%token          RTNYES
%token          ARROW

%type  <number> number
%type  <number> expr

%left '!' '^'
%left '&'
%left '#' '=' '>' '<' GE LE
%left '-' '+'
%left '*' '/'
%left '%'
%left LSHIFT RSHIFT


%%
object:           realobject
                | specialaction object
                ;

realobject:       NAME {
        Entry_Table_iter i = file_entries.find($1);
        Entry* ent = 0;
        if (i != file_entries.end()) {
                ent = i->second;
        } else {
                i = global_entries.find($1);
                if (i == global_entries.end() && !firstpass) {
                        char *rest;
                        unsigned long n = strtoul($1, &rest, 10);
                        if (!*rest && $1[0] != '-') {
                                unsigned long addr = get_address_of("DOBINT");
                                if (addr == 0) {
                                        yyerror("Cannot find bint prolog (DOBINT)", true);
                                }
                                outputreverse(addr, 5);
                                outputreverse(n, 5);
                        } else {
                                if (strspn($1, "0123456789eE+-.") != strlen($1))
                                        yyerror("Cannot find entry", true);
                                else
                                        doreal($1);
                        }
                } else
                        ent = i->second;
        }
        if (ent != 0)
                switch (ent->type()) {
                case T_POINTER:
                        outputreverse(ent->get_address(), 5);
                        break;
                
                case T_FLASHPTR: {
                        unsigned long addr = get_address_of("DOFLASHP");
                        if (addr == 0)
                                yyerror("Cannot find address of flashpointer prolog (DOFLASHP)", true);
                        else
                                outputreverse(addr, 5);
                        Flashptr* fptr = dynamic_cast<Flashptr*>(ent);
                        outputreverse(fptr->bank_num(), 3);
                        outputreverse(fptr->cmd_num(), 4);
                        break;
                }

                case T_ROMPTR: {
                        unsigned long addr = get_address_of("DOROMP");
                        if (addr == 0)
                                yyerror("Cannot find address of rompointer prolog (DOROMP)", true);
                        else
                                outputreverse(addr, 5);
                        Romptr* rptr = dynamic_cast<Romptr*>(ent);
                        outputreverse(rptr->lib_num(), 3);
                        outputreverse(rptr->cmd_num(), 3);
                        break;
                }
                }
}
                | '#' NAME            { doptrbint($2, false); }
                | PTR NAME            { doptrbint($2, true); }
                | ID STRING           { doId($2, true); }
                | LAM STRING          { doId($2, false); }
                | STRING              { dostring($1); }
                | '$' STRING          { dostring($2); }
                | CHR                 { dochar($1); }
                | REAL NAME           { doreal($2); }
                | EXREAL NAME         { doexreal($2); }
                | CMP NAME NAME       { docmp($2, $3); }
                | EXCMP NAME NAME     { doexcmp($2, $3); }
                | DOCOL objects SEMI
                | '{' objects '}'
                | SYMBOL objects SEMI
                | UNIT objects SEMI
                | MATRIX objects SEMI
                | TAG STRING { dotag($2); } object
                | ROMPTR NAME NAME    { doromptr($2, $3); }
                | FPTR NAME NAME      { doflashptr($2, $3); }
                | ACPTR NAME NAME     { doacptr($2, $3); }
                | LENBODYOB NAME NAME { dolenbody($1, $2, $3); }
                | NIBB NAME NAME      { donibb($2, $3); }
                | NIBBHEX NAME        { donibb(0, $2); }
                | ZINT NAME           { dozint($2); }
                | CODE codeinput ENDCODE {
        if (firstpass)
                code_lens.push_front(pc - code_begin);
}
                ;


specialaction:    LABEL NAME {
        if (firstpass)
                file_entries.insert(make_pair(string($2), new Pointer(pc)));
}
                | DEFINE NAME NAME {
        char *rest;
        unsigned long addr = strtoul($3, &rest, 16);
        if (*rest)
                yyerror("Hexa expected");

        Entry_Table_iter i = file_entries.find($2);
        if (i != file_entries.end()) {
                if (i->second->get_address() != addr)
                        yyerror("warning: Redefinition of label; ignored");
        } else
                file_entries.insert(make_pair(string($2), new Pointer(addr)));
}
                | DCCP NAME NAME {
        char *rest;
        unsigned long val = strtoul($2, &rest, 16);
        if (*rest)
                yyerror("Hexa expected");

        Entry_Table_iter i = file_entries.find($3);
        if (i != file_entries.end()) {
                if (i->second->get_address() != constpointer)
                        yyerror("warning: Redefinition of label; ignored");
        } else {
                file_entries.insert(make_pair(string($3), new Pointer(constpointer)));
                constpointer += val;
        }
}
                ;


objects:          /* empty */
                | objects object
                ;


codeinput:        /* empty */
                | codeinput line
                ;

line:             '\n'
                | mnemonics '\n'
                | error '\n'           { yyerrok; }
                ;

mnemonics:        mnemonic
                | mnemonics mnemonic
                ;

mnemonic:         MNEMADDSUB FIELD             { mnemaddsub($1, $2); }
                | MNEMADDSUB '.' FIELD         { mnemaddsub($1, $3); }
                | MNEMADDSUBCON expr FIELD     { mnemaddsubcon($1, $3, $2); }
                | MNEMADDSUBCON expr '.' FIELD { mnemaddsubcon($1, $4, $2); }
                | MNEMZEROETAL FIELD           { mnemzeroetal($1, $2); }
                | MNEMZEROETAL '.' FIELD       { mnemzeroetal($1, $3); }
                | MNEMLOGICAL FIELD            { mnemlogical($1, $2); }
                | MNEMLOGICAL '.' FIELD        { mnemlogical($1, $3); }
                | MNEMSRB FIELD                { mnemsrb($1, $2); }
                | MNEMSRB '.' FIELD            { mnemsrb($1, $3); }
                | MNEMACLD expr                { mnemacld($1, $2); }
                | MNEMMEMORY FIELD             { mnemmemoryf($1, $2); }
                | MNEMMEMORY '.' FIELD         { mnemmemoryf($1, $3); }
                | MNEMMEMORY expr              { mnemmemoryn($1, $2); }
                | MNEMMEMORY '.' expr          { mnemmemoryn($1, $3); }
                | MNEMBIT expr                 { mnembit($1, $2); }
                | MNEMBIT '.' expr             { mnembit($1, $3); }
                | MNEMDATREGADD expr           { mnemdatregadd($1, $2); }
                | MNEMDATREGLD expr            { mnemdatregld($1, (unsigned long)$2); }
                | MNEMSCRATCHREG FIELD         { mnemscratchreg($1, $2); }
                | MNEMSCRATCHREG '.' FIELD     { mnemscratchreg($1, $3); }
                | MNEMPLD expr                 { mnempld($2); }
                | MNEMCPASN expr               { mnemcpasn($1, $2); }
                | MNEMCPASN '.' expr           { mnemcpasn($1, $3); }
                | MNEMSETST expr               { mnemsetst($1, $2); }
                | MNEMSETST '.' expr           { mnemsetst($1, $3); }
                | MNEMGOTO expr                { mnemgoto($1, $2); }
                | MNEMGOTOC expr               {
        output($1 | 0x60, 3);      /* GOC or GONC 6 nibbles futher */
        mnemgoto("GOTO", $2);
}
                | MNEMTEST1 FIELD              { mnemtest1($1, $2); }
                | MNEMTEST1 '.' FIELD          { mnemtest1($1, $3); }
                | MNEMTEST2 FIELD              { mnemtest2($1, $2); }
                | MNEMTEST2 '.' FIELD          { mnemtest2($1, $3); }
                | MNEMTEST3 expr               { mnemtest3($1, $2); }
                | MNEMTEST3 '.' expr           { mnemtest3($1, $3); }
                | MNEMTEST4                    { last_test_opcode = $1; last_test_len = 3; }
                | MNEMTEST5 expr               { mnemtest5($1, $2); }
                | MNEMTEST5 '.' expr           { mnemtest5($1, $3); }
                | GOYES expr                   { goyes($2); }
                | ARROW expr                   { goyes($2); }
                | RTNYES                       {
        if (last_test_len) {
                int len = last_test_len;
                last_test_len = 0;
                output(last_test_opcode, len);
                output(0, 2);
        } else
                yyerror("Must follow a test");
}
                | MNEMNOFIELD2                 { mnemnofield2($1); }
                | MNEMNOFIELD3                 { mnemnofield3($1); }
                | MNEMNOFIELD4                 { mnemnofield4($1); }
                | MNEMNOFIELD5                 { output($1, 5); }
                | DEFINE NAME expr             {
        Entry_Table_iter i = file_entries.find($2);
        if (i != file_entries.end()) {
                if (i->second->get_address() != (unsigned long)$3)
                        yyerror("warning: Redefinition of label; ignored");
        } else
                file_entries.insert(make_pair(string($2), new Pointer($3)));
}
                | DEFINE FIELD expr            { yyerror("Invalid constant name"); }
                | CON expr                     { outputreverse($2, $1); } 
                | HEX_NUMBER {
        for (unsigned int i = 0; i < strlen($1); ++i)
                if ($1[i] >= 'A')
                        output($1[i] - 'A' + 10, 1);
                else
                        output($1[i] - '0', 1);
}
                | HEX_REV_NUMBER {
        for (int i = strlen($1) - 1; i >= 0; --i)
                if ($1[i] >= 'A')
                        output($1[i] - 'A' + 10, 1);
                else
                        output($1[i] - '0', 1);
}
                | GOINAC expr                  {
        long value = $2 - pc;
        if ($1 == 0)
                output(3, 1);
        else
                output(0x8082, 4);

        output(4, 1);
        outputreverse(value, 5);
}
                | GOINN expr                   { outputreverse($2 - pc, $1); }
                | CP expr                      { constpointer = $2; }
                | DCCP expr NAME               {
        Entry_Table_iter i = file_entries.find($3);
        if (i != file_entries.end()) {
                if (i->second->get_address() != constpointer)
                        yyerror("warning: Redefinition of label; ignored");
        } else {
                file_entries.insert(make_pair(string($3), new Pointer(constpointer)));
                constpointer += $2;
        }
}
                | ':' expr { cpointerincr = $2; } constnames ':'
                | EXIT                         { mnemexitorup($1, true); }
                | UP                           { mnemexitorup($1, false); }
                | EXITC                        { mnemexitorup_c($1, true, true); }
                | UPC                          { mnemexitorup_c($1, false, true); }
                | EXITNC                       { mnemexitorup_c($1, true, false); }
                | UPNC                         { mnemexitorup_c($1, false, false); }
                | '{' codeinput '}' { mnemskelse(); }
                | SKIPYES { 
        if (last_test_len == 0)
                yyerror("Forbidden"); } '{' codeinput '}' { mnemskelse(); }
                | ARROW '{' codeinput '}' { mnemskelse(); }
                | SKIP  '{' { mnemskip(false); } codeinput '}'
                | SKUB  '{' { mnemskip(true); } codeinput '}'
                | SKIPC  '{' { mnemskip_c(true); } codeinput '}'
                | SKIPNC  '{' { mnemskip_c(false); } codeinput '}'
                | SKIPL  '{' { mnemskip_l(false); } codeinput '}'
                | SKUBL  '{' { mnemskip_l(true); } codeinput '}'
                | STROBJ expr '{' { mnemstrobj($2); } codeinput '}'
                ;


constnames:       NAME            {
        Entry_Table_iter i = file_entries.find($1);
        if (i != file_entries.end()) {
                if (i->second->get_address() != constpointer)
                        yyerror("warning: Redefinition of label; ignored");
        } else {
                file_entries.insert(make_pair(string($1), new Pointer(constpointer)));
                constpointer += cpointerincr;
        }
}
                | constnames NAME {
        Entry_Table_iter i = file_entries.find($2);
        if (i != file_entries.end()) {
                if (i->second->get_address() != constpointer)
                        yyerror("warning: Redefinition of label; ignored");
        } else {
                file_entries.insert(make_pair(string($2), new Pointer(constpointer)));
                constpointer += cpointerincr;
        }
}
                ;


expr:             number                { $$ = $1; }
                | expr '!' expr         { $$ = $1 | $3; }
                | expr '^' expr         { $$ = ($1 & ~$3) | (~$1 & $3); }
                | expr '&' expr         { $$ = $1 & $3; }
                | expr '#' expr         { $$ = ($1 != $3); }
                | expr '=' expr         { $$ = ($1 == $3); }
                | expr '>' expr         { $$ = ($1 > $3); }
                | expr '<' expr         { $$ = ($1 < $3); }
                | expr GE expr          { $$ = ($1 >= $3); }
                | expr LE expr          { $$ = ($1 <= $3); }
                | expr '+' expr         { $$ = $1 + $3; }
                | expr '-' expr         { $$ = $1 - $3; }
                | expr '*' expr         { $$ = $1 * $3; }
                | expr '/' expr         { $$ = ($3 != 0) ? $1 / $3 : ~0; }
                | expr '%' expr         { $$ = $1 % $3; }
                | expr LSHIFT expr      { $$ = $1 << $3; }
                | expr RSHIFT expr      { $$ = $1 >> $3; }
                | '(' expr ')'          { $$ = $2; }
                ;

number:           NUMBER                { $$ = $1; }
                | HEX_NUMBER            { $$ = strtol($1, NULL, 16); }
                | NAME                  {
        Entry_Table_iter i = file_entries.find($1);
        if (i != file_entries.end()) {
                if (i->second->type() != T_POINTER) {
                        $$ = 0;
                        yyerror("Entry type not allowed here");
                } else
                        $$ = i->second->get_address();
        } else {
                i = global_entries.find($1);
                if (i != global_entries.end()) {
                        if (i->second->type() != T_POINTER) {
                                $$ = 0;
                        yyerror("Entry type not allowed here");
                        } else
                                $$ = i->second->get_address();
                } else {
                        if (!firstpass)
                                yyerror("Undefined label", true);
                        $$ = 0;
                }
        }
      
}
                | '&'                   { $$ = pc; }
                ;
