/* $Id: nib_io.hpp,v 1.4 2001/02/08 18:31:17 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifndef __NIB_IO_H
#define __NIB_IO_H

#include <fstream>


/*********************************
 * nibfstream class declaration *
 *********************************/
class nibfstream : public fstream
{
protected:
  unsigned char nibibuff; // Two-nibble input micro-buffer
  bool ibuff;             // F: No nibbles buffered, T: one nibble buffered
  unsigned char nibobuff; // Two-nibble output micro-buffer
  bool obuff;             // F: No nibbles buffered, T: one nibble buffered
#if 0
  string getTranslationFor(unsigned char c, bool expQuote);
#endif

public:
  // Constructors
  nibfstream(): fstream() { ibuff = obuff = false; }
#ifndef MSDOS
  nibfstream(int fd)  : fstream(fd) { ibuff = obuff = false; }
  nibfstream(const char *name, int mode = ios::in, int prot = 0644)
    : fstream(name, mode, prot) { ibuff = obuff = false; }
#else
  nibfstream(const char *name, int mode = ios::in)
    : fstream(name, mode) { ibuff = obuff = false; } 
#endif
  ~nibfstream() { close(); }

#ifndef MSDOS
  void open(const char *name, int mode, int prot = 0664)
    { fstream::open(name, mode, prot); ibuff = obuff = false; }
#else
  void open(const char *name, int mode)
    { fstream::open(name, mode); ibuff = obuff = false; }  
#endif
  void close() { if (obuff) put(nibobuff); fstream::close(); } 

#if 0
  unsigned char getNibble() { // Note that ibuff is inverted
    return ((ibuff = !ibuff)
	    ? (nibibuff = get()) & 0xF : (nibibuff & 0xF0) >> 4);
  }
  unsigned char getChar() {
    unsigned char r = getNibble();
    return r |= (getNibble() << 4);
  }
  unsigned long get3Nibbles();
  unsigned long get4Nibbles();
  unsigned long get5Nibbles();
  string getString(unsigned long &len);
  string getIdNameForm(bool translate);
#endif
  
  void putNibble(unsigned char c) {
    if ((obuff = !obuff) == false) // Note that obuff has been inverted
      put((c << 4) | nibobuff);
    else
      nibobuff = c;
  }
  void putChar(unsigned char c) {
    putNibble(c & 0xF);
    putNibble((c & 0xF0) >> 4);
  }
  void putString(char *str);
  void putIdNameForm(char *str);
  void put5Nibbles(unsigned long a);
  void putReal(double n);
  void putExReal(double n);
#if 0
  void copyNibblesFrom(nibfstream &fIn, unsigned int len);

  long getPos() { return (long) tellg()*2 - ibuff; }
  void setPos(long pos);
  void skipNibbles(long n) { setPos(getPos() + n); }
#endif
};

#endif // __NIB_IO_H
